"use strict";
this.name = "RoutePlanner_Interface";
this.author = "Alnivel";
this.copyright = "2022 Alnivel";
this.version = "0.2";
this.licence = "CC BY-NC-SA 4.0";

this.$ILib = null;

this.$screens = {};

this.$inverseMapTo = function (map) {
    return Object.keys(map).reduce(function (invMap, key) {
        return invMap[map[key]] = key;
    }.bind(this), {})
}

this.$findKeyByValue = function (dictionary, value) {
    var keys = Object.keys(dictionary);
    for (var i = 0; i < keys.length; i++) {
        var key = keys[i];
        if (dictionary[key] === value)
            return key;
    }
    return undefined;
}

this.$colorToColorSpecifierMap = {
    "Green": "greenColor",
    "Yellow": "yellowColor",
    "Red": "redColor",
    "Blue": "blueColor",
    "Cyan": "cyanColor",
}
this.$colorSpecifierToColorMap = this.$inverseMapTo(this.$colorToColorSpecifierMap)

this.$markerShapes = {
    "Cross": "MARKER_X",
    "Plus": "MARKER_PLUS",
    "Square": "MARKER_SQUARE",
    "Diamond": "MARKER_DIAMOND"
}

this.$configureInterface = function () {
    let interfaceScript = this;
    let mainScript = worldScripts["RoutePlanner"];
    let ILib = this.$ILib = worldScripts["RoutePlanner_InterfaceLib"];

    mainScreen: {
        const mainScreen = this.$screens.main = new ILib.PagedScreenController({
            title: "Route Planning Interface",
            screenID: "RoutePlanner_Main",
            allowInterrupt: true,
            exitScreen: "GUI_SCREEN_INTERFACES"
        });

        mainScreen.addToContent(new ILib.LabelLine("General"));
        mainScreen.addToContent(
            new ILib.CycleLine(
                expandDescription("[RoutePlanner_interface_main_wp-order-mode]"),
                mainScript.$waypointOrderModes,
                mainScript.$currentWaypointOrderMode)
                .onSelected(function (selectedIndex, selectedValue) {
                    mainScript.$currentWaypointOrderMode = selectedValue;
                })
        );

        mainScreen.addToContent(
            new ILib.ButtonLine(expandDescription("[RoutePlanner_interface_main_edit-route]"))
                .onUpdate(function () {
                    return {
                        unselectable: mainScript.$currentWaypointOrderMode === "None"
                    };
                })
                .onSelected(function () {
                    this.$screens.editRouteInstructions.runScreen(mainScript.$currentWaypointOrderMode)
                    return true; // returning true prevent current screen redrawing
                }.bind(this))
        );

        mainScreen.addToContent(
            new ILib.ButtonLine(expandDescription("[RoutePlanner_interface_main_route-statistics]"))
                .onUpdate(function () {
                    return {
                        unselectable: mainScript.$currentWaypointOrderMode === "None"
                    };
                })
                .onSelected(function () {
                    this.$screens.routeStatisticsScreen.runScreen()
                    return true;
                }.bind(this))
        );

        const wpTypes = mainScript.$wpTypes;
        const wpTypesList = Object.keys(wpTypes);
        for (let i = 0; i < wpTypesList.length; i++)
        {
            let typeKey = wpTypesList[i];
            let type = wpTypes[typeKey];

            let labelText = expandDescription(
                "[RoutePlanner_interface_main_route-through-toggler]",
                {waypointType: expandDescription(type.nameKey)});
            mainScreen.addToContent(
                new ILib.ToggleLine(labelText, type.isUsed)
                    .onSelected(function (newState) {
                        type.isUsed = newState;
                    })
            )   
        }        

        mainScreen.addToContent(
            new ILib.ButtonLine(expandDescription("[RoutePlanner_interface_main_appearance-settings]"))
                .onSelected(function () {
                    this.$screens.appearanceSettings.runScreen()
                    return true; // returning true prevent current screen redrawing
                }.bind(this))
        );
    }

    const returnToMain = this.$screens.main.returnToScreen.bind(this.$screens.main);

    appearanceSettingsScreen: {
        const appearanceSettingsScreen = this.$screens.appearanceSettings = new ILib.PagedScreenController({
            title: "Route Planning Interface: Appearance Settings",
            screenID: "RoutePlanner_Main_AppearanceSettings",
            allowInterrupt: true,
            exitScreen: "GUI_SCREEN_INTERFACES"
        }, returnToMain);

        const colors = this.$colorToColorSpecifierMap;
        const appearance = mainScript.$appearance;

        appearanceSettingsScreen.addToContent(
            new ILib.CycleLine(
                expandDescription("[RoutePlanner_interface_appearance_user-wp-marker-shape]"),
                this.$markerShapes,
                this.$findKeyByValue(this.$markerShapes, appearance.userWaypointShape))
                .onSelected(function (selectedShapeIndex, selectedShape) {
                    appearance.userWaypointShape = selectedShape;
                })
        );

        appearanceSettingsScreen.addToContent(
            new ILib.CycleLine(
                expandDescription("[RoutePlanner_interface_appearance_user-wp-marker-color]"),
                colors,
                this.$findKeyByValue(colors, appearance.userWaypointColor))
                .onSelected(function (selectedColorIndex, selectedColor) {
                    appearance.userWaypointColor = selectedColor;
                })
        );

        appearanceSettingsScreen.addToContent(
            new ILib.CycleLine(
                expandDescription("[RoutePlanner_interface_appearance_user-route-color-jumps]"),
                colors,
                this.$findKeyByValue(colors, appearance.routeLineColorByJumps))
                .onSelected(function (selectedColorIndex, selectedColor) {
                    appearance.routeLineColorByJumps = selectedColor;
                })
        );

        appearanceSettingsScreen.addToContent(
            new ILib.CycleLine(
                expandDescription("[RoutePlanner_interface_appearance_user-route-color-time]"),
                colors,
                this.$findKeyByValue(colors, appearance.routeLineColorByTime))
                .onSelected(function (selectedColorIndex, selectedColor) {
                    appearance.routeLineColorByTime = selectedColor;
                })
        );
    }

    function formatWaypoints(waypoints, highlightedSystemId, waypointOrderMode) {
        function getSystemSubstitution(id) {
            if (id < 10) return "%J00" + id;
            if (id < 100) return "%J0" + id;
            else return "%J" + id;
        }

        const wpSubtitutions = waypoints.map(function (wp) {
            if (highlightedSystemId === wp)
                return " \[" + getSystemSubstitution(wp) + "\] ";
            else
                return getSystemSubstitution(wp);
        });

        const wpSeparator = waypointOrderMode === "Auto" ? ", " : " → ";

        return wpSubtitutions.length? expandDescription(wpSubtitutions.join(wpSeparator)) : "None";
    }

    routeEditingInstruction: {
        const instructionSP = {
            title: "Route Planning Interface: Route editing instruction",
            screenID: "RoutePlanner_EditRoute_Instructions",
            allowInterrupt: false,
            exitScreen: "GUI_SCREEN_SHORT_RANGE_CHART",
        }

        instructionSP.choices = {
            "00_SPACER" : { text: "" },
            "01_RETURN": {
                text: ILib.$padTextOnSide(
                    expandDescription("[RoutePlanner_interface_route-editing-instructions_return]"), 
                    "right", 32),
                color: "cyanColor"
            },
            "02_ADD_START": {
                text: ILib.$padTextOnSide(
                    expandDescription("[RoutePlanner_interface_route-editing-instructions_add-start]"), 
                    "right", 32),
                color: "cyanColor"
            },
            "03_CONTINUE": {
                text: ILib.$padTextOnSide(
                    expandDescription("[RoutePlanner_interface_route-editing-instructions_start-edit-on-chart]"), 
                    "right", 32),
                color: "cyanColor"
            },
            "04_SPACER": { text: "" },
            "05_SPACER": { text: "" },
        }

        this.$screens.editRouteInstructions = {
            screenParameters: instructionSP,
            returnToMain: returnToMain,
            userWaypoints: mainScript.$userWaypoints,

            runScreen: function (waypointOrderMode) {
                let targetSystem = player.ship.targetSystem;
                let firstUserWaypoint = this.userWaypoints.length === 0? -1 : this.userWaypoints[0];

                const addToStartChoise = this.screenParameters.choices["02_ADD_START"];
                if(system.ID === targetSystem || firstUserWaypoint === targetSystem)
                {
                    addToStartChoise.unselectable = true;
                    addToStartChoise.color = "darkGrayColor";
                }
                else
                {
                    addToStartChoise.unselectable = false;
                    addToStartChoise.color = "cyanColor";
                }

                let callback = function (choice) {
                    if (choice === "01_RETURN") this.returnToMain();
                    else if (choice === "02_ADD_START") 
                    {                                            
                        this.userWaypoints.unshift(targetSystem);
                        interfaceScript.$screens.editRouteInstructions.runScreen(waypointOrderMode);
                    }
                    else mainScript.$routeEditingEnabled = true;
                };
                
                

                instructionSP.message = expandMissionText("RoutePlanner_interface_route-editing-instructions", {
                    waypointOrderMode: waypointOrderMode,
                    userWaypoints: formatWaypoints(this.userWaypoints, targetSystem, waypointOrderMode)
                });

                mission.runScreen(this.screenParameters, callback, this);
            },


        }
    }

    routeEditingActions: {
        const routeEditActionSP = {
            title: "Route Planning Interface: <systemname>",
            screenID: "RoutePlanner_EditRoute_SystemAction",
            allowInterrupt: false,
            exitScreen: "GUI_SCREEN_SHORT_RANGE_CHART",
            choices: {}
        }

        this.$screens.editRouteSystemAction = {
            screenParameters: routeEditActionSP,
            returnToMain: returnToMain,
            userWaypoints: mainScript.$userWaypoints,

            selectedSystemId: null,
            selectedSystemIndex: null,

            systemId: null,
            systemIndex: null,

            runScreen: function (systemId, waypointOrderMode) {
                const systemIndex = this.userWaypoints.indexOf(systemId);
                const selectedSystemId = this.selectedSystemId;

                this.systemId = systemId;
                this.systemIndex = systemIndex;


                const systemInWaypoints = systemIndex !== -1;
                const systemName = System.systemNameForID(systemId);
                let selectedSystemName = undefined;
                let systemIsSelected = false;

                if (selectedSystemId === null)
                    this.selectedSystemIndex = null;
                else {
                    this.selectedSystemIndex = this.userWaypoints.indexOf(selectedSystemId);
                    selectedSystemName = System.systemNameForID(selectedSystemId);
                    systemIsSelected = systemId === selectedSystemId;
                }

                this.screenParameters.message = this.getText(systemId, waypointOrderMode);
                this.screenParameters.choices = this.getChoices(systemInWaypoints, systemIsSelected, selectedSystemName);

                this.screenParameters.title = "Route Planning Interface: " + systemName;

                mission.runScreen(this.screenParameters, this.screenCallback, this);
            },

            getText: function (highlightedSystemId, waypointOrderMode) {
                

                return expandMissionText("RoutePlanner_interface_route-editing-actions", {
                    waypointOrderMode: waypointOrderMode,
                    userWaypoints: formatWaypoints(this.userWaypoints, highlightedSystemId, waypointOrderMode)
                });
            },

            getChoices: function (systemInWaypoints, systemIsSelected, selectedSystemName) {
                const choices = {};

                if (systemInWaypoints) {
                    if (systemIsSelected)
                        choices["01_DESELECT"] = "Deselect";
                    else
                        choices["01_SELECT"] = "Select";

                    choices["02_REMOVE"] = "Remove";
                    if (selectedSystemName) {
                        choices["03_SWAP"] = "Swap with " + selectedSystemName;
                    }
                    // choices["03_SPACER"] = "";
                }
                else {
                    if (selectedSystemName) {
                        choices["01_ADD_BEFORE"] = "Add before " + selectedSystemName;
                        choices["02_ADD_AFTER"] = "Add after " + selectedSystemName;
                    }
                    else {
                        choices["01_SPACER"] = "";
                        choices["02_SPACER"] = "";
                    }

                    choices["03_ADD_END"] = "Add to the end";
                }

                choices["04_SPACER"] = ""
                choices["05_BACK_TO_CHART"] = "Back to chart";
                choices["06_STOP_EDIT"] = "Stop route editing";

                return choices;
            },

            screenCallback: function (choice) {
                if (choice === "05_BACK_TO_CHART") return;
                else if (choice === "06_STOP_EDIT") {
                    mainScript.$routeEditingEnabled = false;
                    this.returnToMain();
                    return;
                }

                if (this.systemIndex !== -1) {
                    if (choice === "01_SELECT")
                        this.selectedSystemId = this.systemId;
                    else if (choice === "01_DESELECT")
                        this.selectedSystemId = null;
                    else if (choice === "02_REMOVE") {
                        this.userWaypoints.splice(this.systemIndex, 1);
                        if (this.selectedSystemIndex === this.systemIndex)
                            this.selectedSystemId = null
                    }

                    else if (choice === "03_SWAP") {
                        let temp = this.userWaypoints[this.systemIndex];
                        this.userWaypoints[this.systemIndex] = this.userWaypoints[this.selectedSystemIndex];
                        this.userWaypoints[this.selectedSystemIndex] = temp;

                        this.selectedSystemIndex = this.systemIndex;
                    }
                }
                else {
                    if (choice == "01_ADD_BEFORE")
                        this.userWaypoints.splice(this.selectedSystemIndex, 0, this.systemId)
                    else if (choice == "02_ADD_AFTER")
                        this.userWaypoints.splice(this.selectedSystemIndex + 1, 0, this.systemId)
                    else if (choice == "03_ADD_END")
                        this.userWaypoints.push(this.systemId)
                }

            }
        }
    }

    routeStatistics: {
        const routeStatisticsScreen = this.$screens.routeStatisticsScreen = new ILib.PagedScreenController({
            title: "Route Planning Interface: Statistics",
            screenID: "RoutePlanner_Main_RouteStatistics",
            allowInterrupt: true,
            exitScreen: "GUI_SCREEN_INTERFACES"
        }, returnToMain);

        let formatToColumns = function (name, type, remainTime, travelTime, distance){
            return "" +
                ILib.$padTextOnSide(name, "right", 6) + 
                ILib.$padTextAround(type, 8) +
                ILib.$padTextOnSide(remainTime, "left", 6) +
                ILib.$padTextOnSide(travelTime, "left", 6) +
                ILib.$padTextOnSide(distance, "left", 6);
        } 
        
        routeStatisticsScreen.addToHeader(
            new ILib.LabelLine(formatToColumns("Name", "Type", "Remain time", "Travel time", "Distance, LY"))
        );

        routeStatisticsScreen.addToHeader(new ILib.LabelLine(ILib.$repeatPattern("-", 32)));        
        
        let rsScreenRun = routeStatisticsScreen.runScreen.bind(routeStatisticsScreen);
        let updateTable = function (){
            const playerShip = player.ship;
            const targetSystem = playerShip.targetSystem;

            const memoizedRoutes = mainScript.$memoizedRoutes;
            const routeMode = playerShip.routeMode;
            const distanceMeasurement = routeMode === "OPTIMIZED_BY_JUMPS" ? "distance" : "distanceWhenTime";
            const timeMeasurement     = routeMode === "OPTIMIZED_BY_JUMPS" ? "timeWhenDistance" : "time";

            const waypointsOnCurrentRoute = mainScript.$waypointsOnCurrentRoute;
            const waypointsOnCurrentRouteCount = waypointsOnCurrentRoute.length;

            const storedRoutes = mainScript.$storedRoutes;
            const storedRoutesCount = storedRoutes.length;

            
            routeStatisticsScreen.clearContent();

            let totalTime = 0;
            let totalDistance = 0;
            let previousWaypointId = system.ID;   

            for(let i = waypointsOnCurrentRouteCount - 1; i >= 0; i--)
            {
                let waypointInfo = waypointsOnCurrentRoute[i];
                if(waypointInfo === undefined) // if something went wrong then just ignore it and hope for better
                    continue;
                let waypointId = waypointInfo.ID;

                let name = System.systemNameForID(waypointId);
                let costs = memoizedRoutes.routeCost(previousWaypointId, waypointId);
                totalDistance += costs[distanceMeasurement];
                totalTime += costs[timeMeasurement];

                let type = "", remainTime = "";
                if(waypointInfo.type)
                    type = expandDescription(waypointInfo.type.nameKey);
                if(waypointId === targetSystem)
                    type += ", " + expandDescription("[RoutePlanner_wpType_targetSystem]");
                if(waypointInfo.eta)
                    remainTime = ((waypointInfo.eta - clock.seconds) / 3600).toFixed(1);

                routeStatisticsScreen.addToContent(
                    new ILib.LabelLine(formatToColumns(name, type, remainTime, totalTime.toFixed(1), totalDistance.toFixed(1)))
                );

                previousWaypointId = waypointId;
            }   

            if(targetSystem !== previousWaypointId) {
                let name = System.systemNameForID(targetSystem);
                let costs = memoizedRoutes.routeCost(system.ID, targetSystem);
                totalDistance = costs[distanceMeasurement];
                totalTime = costs[timeMeasurement];

                routeStatisticsScreen.addToContent(
                    new ILib.LabelLine(formatToColumns(
                        name, expandDescription("[RoutePlanner_wpType_targetSystem]"), 
                        "", totalTime.toFixed(1), totalDistance.toFixed(1)))
                );

                previousWaypointId = targetSystem;
            }

            for(let i = 0; i < storedRoutesCount; i++)
            {
                let waypointInfo = storedRoutes[i].endWaypointInfo;
                if(waypointInfo === undefined) // if something went wrong then just ignore it and hope for better 2
                    continue;
                let waypointId = waypointInfo.ID;

                let name = System.systemNameForID(waypointId);
                let costs = memoizedRoutes.routeCost(previousWaypointId, waypointId);
                totalDistance += costs[distanceMeasurement];
                totalTime += costs[timeMeasurement];

                let type = "", remainTime = "";
                if(waypointInfo.type)
                    type = expandDescription(waypointInfo.type.nameKey);
                if(waypointInfo.eta)
                    remainTime = ((waypointInfo.eta - clock.seconds) / 3600).toFixed(1);

                routeStatisticsScreen.addToContent(
                    new ILib.LabelLine(formatToColumns(name, type, remainTime, totalTime.toFixed(1), totalDistance.toFixed(1)))
                );

                previousWaypointId = waypointId;
            }
            
            routeStatisticsScreen.clearFooter();

            routeStatisticsScreen.addToFooter(new ILib.LabelLine(ILib.$repeatPattern("-", 32)));  

            routeStatisticsScreen.addToFooter(
                new ILib.LabelLine(formatToColumns("Total", "", "", totalTime.toFixed(1), totalDistance.toFixed(1)))
            );

        }

        
        routeStatisticsScreen.runScreen = function() { updateTable(); rsScreenRun(); }

    }
}
